@@ -3,17 +3,19 @@ Base class for finding modules.
 """
 
 import dis
+import glob
 import imp
+import importlib.machinery
+import importlib.util
 import logging
 import marshal
 import opcode
 import os
-import pkgutil
-import re
 import sys
-import types
+import tokenize
 import zipfile
 
+from cx_Freeze.common import rebuild_code_object
 import cx_Freeze.hooks
 
 BUILD_LIST = opcode.opmap["BUILD_LIST"]
@@ -29,18 +31,6 @@ STORE_OPS = (STORE_NAME, STORE_GLOBAL)
 
 __all__ = [ "Module", "ModuleFinder" ]
 
-try:
-    bytes   # Python >= 2.6
-except NameError:
-    bytes = str
-
-try:
-    source_from_cache = imp.source_from_cache  # Python 3.2 and above
-except AttributeError:
-    def source_from_cache(path):  # Pre PEP 3147 - cache is just .pyc/.pyo
-        assert path.endswith(('.pyc', '.pyo'))
-        return path[:-1]
-
 class ZipModulesCache(object):
     """A cache of module and package locations within zip files."""
     def __init__(self):
@@ -107,10 +97,10 @@ class ZipModulesCache(object):
             if ext not in ('.pyc', '.pyo'):
                 continue
             if '__pycache__' in baseName:
-                if sys.version_info[:2] < (3, 2) \
-                        or not baseName.endswith(imp.get_tag()):
+                if not baseName.endswith(sys.implementation.cache_tag):
                     continue
-                baseName = os.path.splitext(source_from_cache(archiveName))[0]
+                baseName = os.path.splitext(
+                        importlib.util.source_from_cache(archiveName))[0]
             nameparts = baseName.split("/")
             
             if len(nameparts) > 1 and nameparts[-1] == '__init__':
@@ -129,24 +119,28 @@ class ZipModulesCache(object):
 
 class ModuleFinder(object):
 
-    def __init__(self, includeFiles = None, excludes = [], path = None,
-            replacePaths = None):
-        self.includeFiles = includeFiles
-        if includeFiles is None:
-            self.includeFiles = []
-        self.excludes = dict.fromkeys(excludes)
-        self.replacePaths = replacePaths
-        if replacePaths is None:
-            self.replacePaths = []
+    def __init__(self, include_files=None, excludes=None, path=None,
+            replace_paths=None, zip_include_all_packages=False,
+            zip_exclude_packages=None, zip_include_packages=None,
+            constants_module=None, zip_includes=None):
+        self.include_files = include_files or []
+        self.excludes = dict.fromkeys(excludes or [])
+        self.optimizeFlag = 0
         self.path = path or sys.path
+        self.replace_paths = replace_paths or []
+        self.zip_include_all_packages = zip_include_all_packages
+        self.zip_exclude_packages = zip_exclude_packages or []
+        self.zip_include_packages = zip_include_packages or []
+        self.constants_module = constants_module
+        self.zip_includes = zip_includes or []
         self.modules = []
         self.aliases = {}
+        self.exclude_dependent_files = {}
         self._modules = dict.fromkeys(excludes)
-        self._builtinModules = dict.fromkeys(sys.builtin_module_names)
-        self._badModules = {}
+        self._builtin_modules = dict.fromkeys(sys.builtin_module_names)
+        self._bad_modules = {}
         self._zip_modules_cache = ZipModulesCache()
         cx_Freeze.hooks.initialize(self)
-        initialExcludedModules = self.excludes.copy()
         self._AddBaseModules()
 
     def _AddBaseModules(self):
@@ -160,31 +154,36 @@ class ModuleFinder(object):
            """
         self.IncludeModule("traceback")
         self.IncludeModule("warnings")
+        self.IncludeModule("unicodedata")
         self.IncludePackage("encodings")
-        if sys.version_info[0] >= 3:
-            self.IncludeModule("io")
+        self.IncludeModule("io")
         self.IncludeModule("os")
         self.IncludeModule("sys")
         self.IncludeModule("zlib")
-        if sys.version_info[:2] >= (3, 4):
-            # We need this, because collections gets loaded (via traceback),
-            # and a partially frozen package causes problems.
-            self.IncludeModule("collections.abc")
-        if sys.version_info[:2] >= (3,5):
-            self.IncludeModule("importlib.abc")
+        self.IncludeModule("collections.abc")
+        self.IncludeModule("importlib.abc")
 
-    def _AddModule(self, name):
+    def _AddModule(self, name, path=None, file_name=None, parent=None):
         """Add a module to the list of modules but if one is already found,
            then return it instead; this is done so that packages can be
            handled properly."""
         module = self._modules.get(name)
         if module is None:
-            module = self._modules[name] = Module(name)
+            module = self._modules[name] = Module(name, path, file_name,
+                    parent)
             self.modules.append(module)
-            if name in self._badModules:
+            if name in self._bad_modules:
                 logging.debug("Removing module [%s] from list of bad modules",
                         name)
-                del self._badModules[name]
+                del self._bad_modules[name]
+            if self.zip_include_all_packages \
+                    and module.name not in self.zip_exclude_packages \
+                    or module.name in self.zip_include_packages:
+                module.store_in_file_system = False
+        if module.path is None and path is not None:
+            module.path = path
+        if module.file is None and file_name is not None:
+            module.file = file_name
         return module
 
     def _DetermineParent(self, caller):
@@ -200,11 +199,11 @@ class ModuleFinder(object):
            package modules. If the package module has not been completely
            imported yet, defer the import until it has been completely imported
            in order to avoid spurious errors about missing modules."""
-        if packageModule.inImport and caller is not packageModule:
+        if packageModule.in_import and caller is not packageModule:
             deferredImports.append((caller, packageModule, fromList))
         else:
             for name in fromList:
-                if name in packageModule.globalNames:
+                if name in packageModule.global_names:
                     continue
                 subModuleName = "%s.%s" % (packageModule.name, name)
                 self._ImportModule(subModuleName, deferredImports, caller)
@@ -218,7 +217,7 @@ class ModuleFinder(object):
                 # Namespace package (?)
                 module = sys.modules[name]
                 info = ("", "", imp.PKG_DIRECTORY)
-                return None, module.__path__[0], info
+                return None, list(module.__path__)[0], info
 
             # Check for modules in zip files.
             # If a path is a subdirectory within a zip file, we must have
@@ -240,7 +239,7 @@ class ModuleFinder(object):
 
     def _ImportAllSubModules(self, module, deferredImports, recursive = True):
         """Import all sub modules to the given package."""
-        suffixes = [s[0] for s in imp.get_suffixes()]
+        suffixes = importlib.machinery.all_suffixes()
 
         for path in module.path:
             try:
@@ -281,7 +280,7 @@ class ModuleFinder(object):
                     if subModuleName not in self.excludes:
                         raise ImportError("No module named %r" % subModuleName)
                 else:
-                    module.globalNames[name] = None
+                    module.global_names[name] = None
                     if subModule.path and recursive:
                         self._ImportAllSubModules(subModule, deferredImports,
                                 recursive)
@@ -291,7 +290,7 @@ class ModuleFinder(object):
         while deferredImports:
             newDeferredImports = []
             for caller, packageModule, subModuleNames in deferredImports:
-                if packageModule.inImport and skipInImport:
+                if packageModule.in_import and skipInImport:
                     continue
                 self._EnsureFromList(caller, packageModule, subModuleNames,
                         newDeferredImports)
@@ -319,7 +318,7 @@ class ModuleFinder(object):
                 module = self._InternalImportModule(fullName,
                         deferredImports, namespace = namespace)
                 if module is not None:
-                    parent.globalNames[name] = None
+                    parent.global_names[name] = None
                     return module
 
             module = self._InternalImportModule(name,
@@ -349,8 +348,8 @@ class ModuleFinder(object):
             if caller is None:
                 raise ImportError("No module named %r" % name)
             self._RunHook("missing", name, caller)
-            if name not in caller.ignoreNames:
-                callers = self._badModules.setdefault(name, {})
+            if name not in caller.ignore_names:
+                callers = self._bad_modules.setdefault(name, {})
                 callers[caller.name] = None
 
         return module
@@ -365,11 +364,11 @@ class ModuleFinder(object):
         except KeyError:
             pass
 
-        if name in self._builtinModules:
+        if name in self._builtin_modules:
             module = self._AddModule(name)
             logging.debug("Adding module [%s] [C_BUILTIN]", name)
             self._RunHook("load", module.name, module)
-            module.inImport = False
+            module.in_import = False
             return module
 
         pos = name.rfind(".")
@@ -384,8 +383,6 @@ class ModuleFinder(object):
                             namespace = namespace)
             if parentModule is None:
                 return None
-            if namespace:
-                parentModule.ExtendPath()
             path = parentModule.path
             searchName = name[pos + 1:]
 
@@ -395,6 +392,16 @@ class ModuleFinder(object):
             self._modules[name] = module
             return module
 
+        # detect namespace packages
+        try:
+            spec = importlib.util.find_spec(name)
+        except (AttributeError, ModuleNotFoundError, ValueError):
+            spec = None
+        if spec and spec.origin in (None, 'namespace') and \
+                spec.submodule_search_locations:
+            path = list(spec.submodule_search_locations)[0]
+            return self._LoadNamespacePackage(name, path, parentModule)
+        # other modules or packages
         try:
             fp, path, info = self._FindModule(searchName, path, namespace)
             if info[-1] == imp.C_BUILTIN and parentModule is not None:
@@ -414,24 +421,20 @@ class ModuleFinder(object):
         if type == imp.PKG_DIRECTORY:
             return self._LoadPackage(name, path, parent, deferredImports,
                     namespace)
-        module = self._AddModule(name)
-        module.file = path
-        module.parent = parent
+        module = self._AddModule(name, file_name=path, parent=parent)
 
         if type == imp.PY_SOURCE:
             logging.debug("Adding module [%s] [PY_SOURCE]", name)
             # Load & compile Python source code
-            if sys.version_info[0] >= 3:
-                # For Python 3, read the file with the correct encoding
-                import tokenize
-                fp = open(path, "rb")
-                encoding, lines = tokenize.detect_encoding(fp.readline)
-                fp = open(path, "U", encoding = encoding)
+            fp = open(path, "rb")
+            encoding, lines = tokenize.detect_encoding(fp.readline)
+            fp = open(path, "U", encoding = encoding)
             codeString = fp.read()
             if codeString and codeString[-1] != "\n":
                 codeString = codeString + "\n"
             try:
-                module.code = compile(codeString, path, "exec")
+                module.code = compile(codeString, path, "exec",
+                                      optimize=self.optimizeFlag)
             except SyntaxError:
                 raise ImportError("Invalid syntax in %s" % path)
         
@@ -444,26 +447,22 @@ class ModuleFinder(object):
                 magic = fp.read(4)
             if magic != imp.get_magic():
                 raise ImportError("Bad magic number in %s" % path)
-            skip_bytes = 8 if (sys.version_info[:2] >= (3,3)) else 4
+            skip_bytes = 8
             if isinstance(fp, bytes):
                 module.code = marshal.loads(fp[skip_bytes+4:])
-                module.inZipFile = True
+                module.source_is_zip_file = True
             else:
                 fp.read(skip_bytes)
                 module.code = marshal.load(fp)
         
         elif type == imp.C_EXTENSION:
             logging.debug("Adding module [%s] [C_EXTENSION]", name)
-            if parent is None:
-                # Our extension loader (see the freezer module) uses imp to
-                # load compiled extensions.
-                self.IncludeModule("imp")
 
         # If there's a custom hook for this module, run it.
         self._RunHook("load", module.name, module)
         
         if module.code is not None:
-            if self.replacePaths:
+            if self.replace_paths:
                 topLevelModule = module
                 while topLevelModule.parent is not None:
                     topLevelModule = topLevelModule.parent
@@ -473,13 +472,23 @@ class ModuleFinder(object):
             # Scan the module code for import statements
             self._ScanCode(module.code, module, deferredImports)
         
-        module.inImport = False
+            # Verify __package__ in use
+            self._ReplacePackageInCode(module)
+
+        module.in_import = False
+        return module
+
+    def _LoadNamespacePackage(self, name, path, parent):
+        """Load the namespace package, given its name and path."""
+        module = self._AddModule(name, path=[path], parent=parent)
+        filename = os.path.join(path, "__init__.py")
+        module.code = compile("", filename, "exec")
+        logging.debug("Adding module [%s] [PKG_NAMESPACE_DIRECTORY]", name)
         return module
 
     def _LoadPackage(self, name, path, parent, deferredImports, namespace):
         """Load the package, given its name and path."""
-        module = self._AddModule(name)
-        module.path = [path]
+        module = self._AddModule(name, path=[path], parent=parent)
         try:
             fp, path, info = self._FindModule("__init__", module.path, False)
             self._LoadModule(name, fp, path, info, deferredImports, parent)
@@ -492,12 +501,41 @@ class ModuleFinder(object):
             logging.debug("Adding module [%s] [PKG_NAMESPACE_DIRECTORY]", name)
         return module
 
+    def _ReplacePackageInCode(self, module):
+        """Replace the value of __package__ directly in the code,
+           only in zipped modules."""
+        co = module.code
+        if co is None or module.parent is None or \
+                module.WillBeStoredInFileSystem() or \
+                "__package__" in module.global_names:
+            # In some modules, like 'six' the variable is defined, so...
+            return
+        # Only if the code references it.
+        if "__package__" in co.co_names:
+            # Insert a bytecode to represent the code:
+            # __package__ = module.parent.name
+            constants = list(co.co_consts)
+            pkg_const_index = len(constants)
+            pkg_name_index = co.co_names.index("__package__")
+            if pkg_const_index > 255 or pkg_name_index > 255:
+                # Don't touch modules with many constants or names;
+                # This is good for now.
+                return
+            # The bytecode/wordcode
+            codes = [LOAD_CONST, pkg_const_index,
+                     STORE_NAME, pkg_name_index]
+            asm_code = bytes(codes)
+            new_code = asm_code + co.co_code
+            constants.append(module.parent.name)
+            code = rebuild_code_object(co, code=new_code, constants=constants)
+            module.code = code
+
     def _ReplacePathsInCode(self, topLevelModule, co):
         """Replace paths in the code as directed, returning a new code object
            with the modified paths in place."""
         # Prepare the new filename.
         origFileName = newFileName = os.path.normpath(co.co_filename)
-        for searchValue, replaceValue in self.replacePaths:
+        for searchValue, replaceValue in self.replace_paths:
             if searchValue == "*":
                 searchValue = os.path.dirname(topLevelModule.file)
                 if topLevelModule.path:
@@ -515,18 +553,7 @@ class ModuleFinder(object):
             if isinstance(value, type(co)):
                 constants[i] = self._ReplacePathsInCode(topLevelModule, value)
         
-        # Build the new code object.
-        if sys.version_info[0] < 3:
-            return types.CodeType(co.co_argcount, co.co_nlocals,
-                    co.co_stacksize, co.co_flags, co.co_code, tuple(constants),
-                    co.co_names, co.co_varnames, newFileName, co.co_name,
-                    co.co_firstlineno, co.co_lnotab, co.co_freevars,
-                    co.co_cellvars)
-        return types.CodeType(co.co_argcount, co.co_kwonlyargcount,
-                co.co_nlocals, co.co_stacksize, co.co_flags, co.co_code,
-                tuple(constants), co.co_names, co.co_varnames, newFileName,
-                co.co_name, co.co_firstlineno, co.co_lnotab, co.co_freevars,
-                co.co_cellvars)
+        return rebuild_code_object(co, constants=constants, filename=newFileName)
 
     def _RunHook(self, hookName, moduleName, *args):
         """Run hook for the given module if one is present."""
@@ -541,9 +568,7 @@ class ModuleFinder(object):
            modules are truly missing."""
         arguments = []
         importedModule = None
-        method = dis._unpack_opargs if sys.version_info[:2] >= (3, 5) \
-                else self._UnpackOpArgs
-        for opIndex, op, opArg in method(co.co_code):
+        for opIndex, op, opArg in dis._unpack_opargs(co.co_code):
 
             # keep track of constants (these are used for importing)
             # immediately restart loop so arguments are retained
@@ -558,8 +583,8 @@ class ModuleFinder(object):
                     relativeImportIndex, fromList = arguments[-2:]
                 else:
                     relativeImportIndex = -1
-                    fromList, = arguments
-                if name not in module.excludeNames:
+                    fromList = arguments[0] if arguments else []
+                if name not in module.exclude_names:
                     importedModule = self._ImportModule(name, deferredImports,
                             module, relativeImportIndex)
                     if importedModule is not None:
@@ -570,12 +595,12 @@ class ModuleFinder(object):
 
             # import * statement: copy all global names
             elif op == IMPORT_STAR and topLevel and importedModule is not None:
-                module.globalNames.update(importedModule.globalNames)
+                module.global_names.update(importedModule.global_names)
 
             # store operation: track only top level
             elif topLevel and op in STORE_OPS:
                 name = co.co_names[opArg]
-                module.globalNames[name] = None
+                module.global_names[name] = None
 
             # reset arguments; these are only needed for import statements so
             # ignore them in all other cases!
@@ -587,36 +612,18 @@ class ModuleFinder(object):
                 self._ScanCode(constant, module, deferredImports,
                         topLevel = False)
 
-    def _UnpackOpArgs(self, code):
-        """Unpack the operations and arguments from the byte code. From Python
-           3.5 onwards this is found in the private method _unpack_opargs
-           but for earlier releases this wasn't available as a separate
-           method."""
-        opIndex = 0
-        numOps = len(code)
-        is3 = sys.version_info[0] >= 3
-        while opIndex < numOps:
-            offset = opIndex
-            if is3:
-                op = code[opIndex]
-            else:
-                op = ord(code[opIndex])
-            opIndex += 1
-            arg = None
-            if op >= dis.HAVE_ARGUMENT:
-                if is3:
-                    arg = code[opIndex] + code[opIndex + 1] * 256
-                else:
-                    arg = ord(code[opIndex]) + ord(code[opIndex + 1]) * 256
-                opIndex += 2
-            yield (offset, op, arg)
-
     def AddAlias(self, name, aliasFor):
         """Add an alias for a particular module; when an attempt is made to
            import a module using the alias name, import the actual name
            instead."""
         self.aliases[name] = aliasFor
 
+    def AddConstant(self, name, value):
+        self.constants_module.values[name] = value
+
+    def ExcludeDependentFiles(self, fileName):
+        self.exclude_dependent_files[fileName] = None
+
     def ExcludeModule(self, name):
         """Exclude the named module from the resulting frozen executable."""
         self.excludes[name] = None
@@ -629,14 +636,15 @@ class ModuleFinder(object):
             moduleName = name
         info = (ext, "r", imp.PY_SOURCE)
         deferredImports = []
-        module = self._LoadModule(moduleName, open(path, "U"), path, info,
-                deferredImports)
+        module = self._LoadModule(moduleName, None, path, info, deferredImports)
         self._ImportDeferredImports(deferredImports)
         return module
 
-    def IncludeFiles(self, sourcePath, targetPath):
+    def IncludeFiles(self, sourcePath, targetPath, copyDependentFiles = True):
         """Include the files in the given directory in the target build."""
-        self.includeFiles.append((sourcePath, targetPath))
+        self.include_files.append((sourcePath, targetPath))
+        if not copyDependentFiles:
+            self.ExcludeDependentFiles(sourcePath)
 
     def IncludeModule(self, name, namespace = False):
         """Include the named module in the frozen executable."""
@@ -658,12 +666,12 @@ class ModuleFinder(object):
 
     def ReportMissingModules(self):
         """Display a list of modules that weren't found."""
-        if self._badModules:
+        if self._bad_modules:
             sys.stdout.write("Missing modules:\n")
-            names = list(self._badModules.keys())
+            names = list(self._bad_modules.keys())
             names.sort()
             for name in names:
-                callers = list(self._badModules[name].keys())
+                callers = list(self._bad_modules[name].keys())
                 callers.sort()
                 sys.stdout.write("? %s imported from %s\n" % \
                         (name, ", ".join(callers)))
@@ -671,20 +679,54 @@ class ModuleFinder(object):
                              "may not be needed on this platform.\n")
             sys.stdout.write("\n")
 
+    def SetOptimizeFlag(self, optimizeFlag):
+        """Set a new value of optimize flag and returns the previous value."""
+        previous = self.optimizeFlag
+        # The value of optimizeFlag is propagated according to the user's
+        # choice and checked in dist.py or main,py. This value is unlikely
+        # to be wrong, yet we check and ignore any divergent value.
+        if -1 <= optimizeFlag <= 2:
+            self.optimizeFlag = optimizeFlag
+        return previous
+
+    def ZipIncludeFiles(self, sourcePath, targetPath):
+        """Include the file(s) in the library.zip"""
+        self.zip_includes.append((sourcePath, targetPath))
+
 
 class Module(object):
 
-    def __init__(self, name):
+    def __init__(self, name, path=None, file_name=None, parent=None):
         self.name = name
-        self.file = None
-        self.path = None
+        self.file = file_name
+        self.path = path
         self.code = None
-        self.parent = None
-        self.globalNames = {}
-        self.excludeNames = {}
-        self.ignoreNames = {}
-        self.inZipFile = False
-        self.inImport = True
+        self.parent = parent
+        self.global_names = {}
+        self.exclude_names = {}
+        self.ignore_names = {}
+        self.source_is_zip_file = False
+        self.in_import = True
+        self.store_in_file_system = True
+        self.dist_files = None
+        if file_name is not None:
+            dir_name = os.path.dirname(file_name)
+            search_path = os.path.join(dir_name, name+"-*.dist-info")
+        elif path:
+            search_path = path[0]+"-*.dist-info"
+        else:
+            search_path = None
+        if search_path:
+            pathnames = glob.glob(search_path)
+            if pathnames:
+                dist_path = pathnames[0]
+                arc_path = os.path.basename(dist_path)
+                # cache file names to use in write modules
+                dist_files = []
+                for fname in os.listdir(dist_path):
+                    dist_files.append((os.path.join(dist_path, fname),
+                                       os.path.join(arc_path, fname)))
+                self.dist_files = dist_files
 
     def __repr__(self):
         parts = ["name=%s" % repr(self.name)]
@@ -695,16 +737,17 @@ class Module(object):
         return "<Module %s>" % ", ".join(parts)
 
     def AddGlobalName(self, name):
-        self.globalNames[name] = None
+        self.global_names[name] = None
 
     def ExcludeName(self, name):
-        self.excludeNames[name] = None
-
-    def ExtendPath(self):
-        self.path = pkgutil.extend_path(self.path, self.name)
-        if self.parent is not None:
-            self.parent.ExtendPath()
+        self.exclude_names[name] = None
 
     def IgnoreName(self, name):
-        self.ignoreNames[name] = None
+        self.ignore_names[name] = None
 
+    def WillBeStoredInFileSystem(self):
+        if self.parent is not None:
+            return self.parent.WillBeStoredInFileSystem()
+        if self.path is None or self.file is None:
+            return False
+        return self.store_in_file_system
